#!/usr/bin/tclsh 
# $Id: dcdot.tcl,v 1.5 2003/01/13 13:59:07 adam Exp $
#

proc RobotRestart {} {
    global robotMoreWork

    set robotMoreWork 0
}

proc RobotTextHtml {url} {
    global URL
   
    set b $URL($url,buf)
    set e {<meta[^>]*>}
    catch {unset $URL($url,meta)}
    while {[regexp -nocase -indices $e $b i]} {
        set meta [string range $b [lindex $i 0] [lindex $i 1]]
	lappend URL($url,meta) $meta
        set b [string range $b [lindex $i 1] end]
    }
    set b $URL($url,buf)
    set e {<title>[^>]*>}
    catch {unset $URL($url,meta)}
    while {[regexp -nocase -indices $e $b i]} {
        set title [string range $b [lindex $i 0] [lindex $i 1]]
	lappend URL($url,title) $title
        set b [string range $b [lindex $i 1] end]
    }
}

proc Robot200 {url} {
    global URL domains
    
    # puts "Parsing $url"
    switch $URL($url,head,content-type) {
	text/html {
	    RobotTextHtml $url
	}
    }
    # puts "Parsing done"
}

proc RobotReadContent {url sock} {
    global URL

    set buffer [read $sock 16384]
    set readCount [string length $buffer]
    
    if {$readCount <= 0} {
	close $sock
	Robot200 $url
	RobotRestart
    } else {
	# puts "Got $readCount bytes"
	set URL($url,buf) $URL($url,buf)$buffer
    }
}

proc RobotReadHeader {url sock} {
    global URL

    set buffer [read $sock 2148]
    set readCount [string length $buffer]
    
    if {$readCount <= 0} {
	close $sock
	RobotRestart
    } else {
	# puts "Got $readCount bytes"
	set URL($url,buf) $URL($url,buf)$buffer
	
	set n [string first \r\n\r\n $URL($url,buf)]
	if {$n > 1} {
	    puts "string first match n = $n"
	    set code 0
	    set version {}
	    set headbuf [string range $URL($url,buf) 0 $n]
	    incr n 4
	    set URL($url,buf) [string range $URL($url,buf) $n end]
	    
	    regexp {^HTTP/([0-9.]+)[ ]+([0-9]+)} $headbuf x version code
	    set lines [split $headbuf \n]
	    foreach line $lines {
		if {[regexp {^([^:]+):[ ]+([^;]*)} $line x name value]} {
		    set URL($url,head,[string tolower $name]) [string trim $value]
		}
	    }
	    set URL($url,state) skip
            puts "code=$code"	
	    switch $code {
		200 {
		    if {![info exists URL($url,head,content-type)]} {
			set URL($url,head,content-type) {}
		    }
		    switch $URL($url,head,content-type) {
			text/html {
			    fileevent $sock readable [list RobotReadContent $url $sock]
			}
			text/plain {
			    fileevent $sock readable [list RobotReadContent $url $sock]
			}
			application/pdf {
			    puts "ok preceeed with this thingy"
			    fileevent $sock readable [list RobotReadContent $url $sock]
			}
			default {
			    close $sock
			    Robot200 $url
			    RobotRestart
			}
		    }
		}
		default {
		    close $sock
		    RobotRestart
		}
	    }
	}
    }
}

proc RobotConnect {url sock} {
    global URL agent

    fconfigure $sock -translation {lf crlf} -blocking 0
    fileevent $sock readable [list RobotReadHeader $url $sock]
    puts $sock "GET $URL($url,path) HTTP/1.0"
    puts $sock "Host: $URL($url,host)"
    puts $sock "User-Agent: $agent"
    puts $sock ""
    flush $sock
}

proc RobotGetUrl {url phost} {
    global URL
    if {![regexp {([^:]+)://([^/]+)([^ ]*)} $url x method hostport path]} {
        return -1
    }
    if {![regexp {([^:]+):([0-9]+)} $hostport x host port]} {
	set port 80
	set host $hostport
    }
    set URL($url,method) $method
    set URL($url,host) $host
    set URL($url,port) $port
    set URL($url,path) $path
    set URL($url,state) head
    set URL($url,buf) {}
    if [catch {set sock [socket -async $host $port]}] {
        return -1
    }
    RobotConnect $url $sock

    return 0
}

set agent "dcdot.tcl/0.0"
if {![catch {set os [exec uname -s -r]}]} {
    set agent "$agent ($os)"
}

proc RobotGetDCDOT {url} {
    global robotMoreWork 1

    set robotMoreWork 1
    if [RobotGetUrl $url {}] {
	set robotMoreWork 0
    }

    while {$robotMoreWork} {
	vwait robotMoreWork
    }
}

if {$argc == 1} {
    set url [lindex $argv 0]
    RobotGetDCDOT $url
    set mask {,meta}
    if {[info exist URL($url,meta)]} {
        foreach m $URL($url,meta) {
	    puts $m
        }
    }
    if {[info exist URL($url,title)]} {
        foreach m $URL($url,title) {
	    puts $m
        }
    }
    foreach v [array names URL $url,head,*] {
	puts "$v = $URL($v)"
    }
    puts "Buffer length is [string length $URL($url,buf)]"
    set f [open out.pdf w]
    puts -nonewline $f $URL($url,buf)
    close $f
}
